@@ -5,6 +5,9 @@ import sys
 import collections
 import inspect
 from datetime import datetime
+from packaging import version
+from torch.utils.data import Dataset, IterableDataset
+from tqdm import tqdm
 import warnings
 import os
 import random
@@ -25,7 +28,6 @@ except ImportError:
 class Progbar(object):
     """进度条，直接从keras引入
     """
-
     def __init__(self, target, width=30, verbose=1, interval=0.05, stateful_metrics=None):
         self.target = target
         self.width = width
@@ -60,6 +62,7 @@ class Progbar(object):
                 # means "take an average from a single value" but keeps the
                 # numeric formatting.
                 self._values[k] = [v, 1]
+        
         self._seen_so_far = current
 
         now = time.time()
@@ -178,10 +181,18 @@ class CallbackList(object):
         for callback in self.callbacks:
             callback.set_params(params)
 
+    def set_trainer(self, trainer):
+        for callback in self.callbacks:
+            callback.set_trainer(trainer)
+
     def set_model(self, model):
         for callback in self.callbacks:
             callback.set_model(model)
 
+    def set_optimizer(self, optimizer):
+        for callback in self.callbacks:
+            callback.set_optimizer(optimizer)
+
     def on_epoch_begin(self, global_step, epoch, logs=None):
         # 如果是分布式DDP训练，则仅masker_rank可以callback
         if (self.master_rank is not None) and (self.master_rank!=torch.distributed.get_rank()):
@@ -264,11 +275,17 @@ class Callback(object):
     '''Callback基类
     '''
     def __init__(self):
-        self.model = None
+        self.trainer = None  # trainer
+        self.model = None  # nn.Module模型，或者包含Trainer的nn.Module
+        self.optimizer = None  # 优化器
     def set_params(self, params):
         self.params = params
+    def set_trainer(self, trainer):
+        self.trainer = trainer
     def set_model(self, model):
         self.model = model
+    def set_optimizer(self, optimizer):
+        self.optimizer = optimizer
     def on_train_begin(self, logs=None):
         pass
     def on_train_end(self, logs=None):
@@ -289,8 +306,9 @@ class Callback(object):
 
 class BaseLogger(Callback):
     """计算metrics的均值, 默认是callbacks中第一项
+    
+    :param stateful_metrics: List[str], 仅保留状态信息的指标
     """
-
     def __init__(self, stateful_metrics=None):
         if stateful_metrics:
             self.stateful_metrics = set(stateful_metrics)
@@ -337,7 +355,7 @@ class TerminateOnNaN(Callback):
         if loss is not None:
             if np.isnan(loss) or np.isinf(loss):
                 print('Step %d: Invalid loss, terminating training' % global_step)
-                self.model.stop_training = True
+                self.trainer.stop_training = True
 
 
 class ProgbarLogger(Callback):
@@ -412,6 +430,78 @@ class ProgbarLogger(Callback):
             print('%s - Finish Training' % (time_start))
 
 
+class TqdmProgressBar(ProgbarLogger):
+    """ Tqdm进度条
+    """
+    def on_epoch_begin(self, global_step=None, epoch=None, logs=None):
+        if self.verbose:
+            time_start = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
+            print('%s - Epoch: %d/%d' % (time_start, epoch+1, self.epochs))
+            self.target = self.params['steps']
+            self.progbar = tqdm(total=self.params['steps'], desc='Training', dynamic_ncols=False, file=sys.stdout, smoothing=0)
+        self.seen = 0
+        self._values = collections.OrderedDict()
+        self._seen_so_far = 0
+
+    def on_batch_begin(self, global_step=None, local_step=None, logs=None):
+        if self.seen < self.target:
+            self.log_values = []
+
+    def on_batch_end(self, global_step=None, local_step=None, logs=None):
+        self.seen += 1
+        logs = self.smooth_values(self.seen, logs or {})
+        for k in self.params['metrics']:
+            if k in logs:
+                self.log_values.append((k, logs[k]))
+
+        # Skip progbar update for the last batch;
+        # will be handled by on_epoch_end.
+        if self.verbose and self.seen < self.target:
+            self.progbar.n = self.seen
+            self.progbar.refresh()
+            self.progbar.set_postfix(self.log_values)
+
+    def on_epoch_end(self, global_step=None, epoch=None, logs=None):
+        logs = self.smooth_values(self.seen, logs or {})
+        for k in self.params['metrics']:
+            if k in logs:
+                self.log_values.append((k, logs[k]))
+        if self.verbose:
+            self.progbar.n = self.seen
+            self.progbar.refresh()
+            self.progbar.set_postfix(self.log_values)
+            self.progbar.close()
+    
+    def smooth_values(self, current, values=None):
+        '''从Progbar迁移过来
+        '''
+        values = values or []
+        for k, v in values.items():
+            if k not in self.stateful_metrics:
+                if k not in self._values:
+                    self._values[k] = [v * (current - self._seen_so_far), current - self._seen_so_far]
+                else:
+                    self._values[k][0] += v * (current - self._seen_so_far)
+                    self._values[k][1] += (current - self._seen_so_far)
+            else:
+                self._values[k] = [v, 1]
+        self._seen_so_far = current
+
+        logs = collections.OrderedDict()
+        for k in self._values:
+            if isinstance(self._values[k], list):
+                avg = np.mean(
+                    self._values[k][0] / max(1, self._values[k][1]))
+                if abs(avg) > 1e-3:
+                    logs[k] = ' %.4f' % avg
+                else:
+                    logs[k] = ' %.4e' % avg
+            else:
+                logs[k] = ' %s' % self._values[k]
+
+        return logs
+
+
 class History(Callback):
     """指标历史，默认是fit的返回项, 这里仅记录epoch_end的指标
     """
@@ -429,6 +519,15 @@ class History(Callback):
 class EarlyStopping(Callback):
     '''Stop training策略, 从keras中移植
        使用时候需要保证monitor在logs中，因此如果以valid指标评估需要在EarlyStopping前进行评估并设置logs[monitor]
+
+       :param monitor: str, 监控指标，需要在logs中，默认为'loss', 
+       :param min_delta: float, 最小变动，默认为0 
+       :param patience: int, 最长等候的次数，默认为0
+       :param verbose: int, 是否打印，默认为0表示不打印
+       :param mode: str, 控制监控指标monitor的大小方向，默认为'auto', 可选{'auto', 'min', 'max'}
+       :param method: str, 控制是按照epoch还是step来计算，默认为'epoch', 可选{'step', 'epoch'}
+       :param baseline: None/float, 基线, 默认为None 
+       :param restore_best_weights: bool, stopping时候是否恢复最优的权重，默认为False
     '''
     def __init__(self, monitor='loss', min_delta=0, patience=0, verbose=0, mode='auto', method='epoch', baseline=None, restore_best_weights=False):
         super(EarlyStopping, self).__init__()
@@ -488,7 +587,7 @@ class EarlyStopping(Callback):
             self.wait += 1
             if self.wait >= self.patience:
                 self.stopped_iteration = iteration
-                self.model.stop_training = True
+                self.trainer.stop_training = True
                 # 恢复最优权重
                 if self.restore_best_weights:
                     if self.verbose > 0:
@@ -509,6 +608,16 @@ class EarlyStopping(Callback):
 
 class ReduceLROnPlateau(Callback):
     """当monitor指标不下降时候，降低学习率
+
+    :param monitor: str, 监控指标，需要在logs中，默认为'loss'
+    :param factor: float, 权重衰减系数，取值范围(0, 1)，默认为0.1
+    :param patience: int, 最长等候的次数，默认为0
+    :param method: str, 控制是按照epoch还是step来计算，默认为'epoch', 可选{'step', 'epoch'}
+    :param verbose: int, 是否打印，默认为0表示不打印
+    :param mode: str, 控制监控指标monitor的大小方向，默认为'auto', 可选{'auto', 'min', 'max'}
+    :param min_delta: float, 最小变动，默认为0 
+    :param cooldown: float
+    :param min_lr: float, 最小学习率
     """
     def __init__(self, monitor='loss', factor=0.1, patience=10, method='epoch', 
                  verbose=0, mode='auto', min_delta=1e-4, cooldown=0, min_lr=0,
@@ -564,7 +673,7 @@ class ReduceLROnPlateau(Callback):
 
     def process(self, iteration, logs=None):
         logs = logs or {}
-        for i, params in enumerate(self.model.optimizer.param_groups):
+        for i, params in enumerate(self.optimizer.param_groups):
             if i == 0:
                 logs['lr'] = params["lr"]  # 默认第一个param_group作为lr
             else:
@@ -586,7 +695,7 @@ class ReduceLROnPlateau(Callback):
             elif not self.in_cooldown():
                 self.wait += 1
                 if self.wait >= self.patience:
-                    for params in self.model.optimizer.param_groups:
+                    for params in self.optimizer.param_groups:
                         if 'lr' not in params:
                             continue
                         old_lr = float(params["lr"])
@@ -605,6 +714,12 @@ class ReduceLROnPlateau(Callback):
         
 class RemoteMonitor(Callback):
     """Callback used to stream events to a server.
+
+    :param root: str, url+port
+    :param path: str, router
+    :param field: str, 字段名
+    :param headers: header
+    :param send_as_json: bool, 是否以json形式请求，默认为False
     """
     def __init__(self, root='http://localhost:9000', path='/publish/epoch/end/', field='data',
                  headers=None, send_as_json=False):
@@ -634,6 +749,12 @@ class RemoteMonitor(Callback):
 
 class Checkpoint(Callback):
     '''保存Checkpoint, 可以每个epoch或者每隔一定的steps保存
+
+    :param checkpoint_path: str, 模型保存路径(含文件名)，可以使用{epoch}和{step}占位符
+    :param optimizer_path: str, 优化器保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
+    :param steps_params_path: str, 模型训练进度保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
+    :param method: str, 按照轮次保存还是按照步数保存，默认为'epoch'表示每个epoch保存一次, 可选['epoch', 'step'] 
+    :param step_interval: int, method设置为'step'时候指定每隔多少步数保存模型，默认为100表示每隔100步保存一次
     '''
     def __init__(self, checkpoint_path, optimizer_path=None, steps_params_path=None, method='epoch', step_interval=100):
         super().__init__()
@@ -664,16 +785,26 @@ class Checkpoint(Callback):
             filepath = self.optimizer_path.format(epoch=suffix, **logs) if self.method == 'epoch' else self.optimizer_path.format(step=suffix, **logs)
             save_dir = os.path.dirname(filepath)
             os.makedirs(save_dir, exist_ok=True)
-            torch.save(self.model.optimizer.state_dict(), filepath)
+            torch.save(self.optimizer.state_dict(), filepath)
         if self.steps_params_path:
             filepath = self.steps_params_path.format(epoch=suffix, **logs) if self.method == 'epoch' else self.steps_params_path.format(step=suffix, **logs)
             save_dir = os.path.dirname(filepath)
             os.makedirs(save_dir, exist_ok=True)
-            self.model.save_steps_params(filepath)
+            self.trainer.save_steps_params(filepath)
 
 
 class Evaluator(Checkpoint):
     '''评估并保存最优Checkpoint, 也可以只评估
+
+    :param monitor: str, 监控指标，需要在logs中，默认为'perf'
+    :param verbose: int, 是否打印，默认为1表示打印
+    :param mode: str, 控制监控指标monitor的大小方向，默认为'auto', 可选{'auto', 'min', 'max'}
+    :param method: str, 控制是按照epoch还是step来计算，默认为'epoch', 可选{'step', 'epoch'}
+    :param baseline: None/float, 基线, 默认为None 
+    :param checkpoint_path: str, 模型保存路径(含文件名)，可以使用{epoch}和{step}占位符
+    :param optimizer_path: str, 优化器保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
+    :param steps_params_path: str, 模型训练进度保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
+    :param step_interval: int, method设置为'step'时候指定每隔多少步数保存模型，默认为100表示每隔100步保存一次
     '''
     def __init__(self, monitor='perf', mode='max', verbose=1, checkpoint_path=None, optimizer_path=None, steps_params_path=None, method='epoch', step_interval=100):
         super().__init__(checkpoint_path, optimizer_path, steps_params_path, method, step_interval)
@@ -709,6 +840,13 @@ class Logger(Callback):
     '''默认logging
     对于valid/dev和test的日志需要在evaluate之后对log进行赋值，如log['dev_f1']=f1，并在Evaluator之后调用
     若每隔一定steps对验证集评估，则Logger的interval设置成和Evaluater一致或者约数，保证日志能记录到
+
+    :param log_path: str, log文件的保存路径
+    :param interval: int, 保存log的间隔
+    :param mode: str, log保存的模式, 默认为'a'表示追加
+    :param separator: str, 指标间分隔符
+    :param verbosity: int, 可选[0,1,2]，指定log的level
+    :param name: str, 默认为None
     '''
     def __init__(self, log_path, interval=10, mode='a', separator='\t', verbosity=1, name=None):
         super(Logger, self).__init__()
@@ -750,6 +888,12 @@ class Tensorboard(Callback):
     对于valid/dev和test的Tensorboard需要在evaluate之后对log进行赋值，如log['dev/f1']=f1，并在Evaluator之后调用
     赋值需要分栏目的用'/'进行分隔
     若每隔一定steps对验证集评估，则Tensorboard的interval设置成和Evaluater一致或者约数，保证Tensorboard能记录到
+
+    :param log_dir: str, tensorboard文件的保存路径
+    :param method: str, 控制是按照epoch还是step来计算，默认为'epoch', 可选{'step', 'epoch'}
+    :param interval: int, 保存tensorboard的间隔
+    :param prefix: str, tensorboard分栏的前缀，默认为'train'
+    :param on_epoch_end_scalar_epoch: bool, epoch结束后是横轴是按照epoch还是global_step来记录
     '''
     def __init__(self, log_dir, method='epoch', interval=10, prefix='train', on_epoch_end_scalar_epoch=True):
         super(Tensorboard, self).__init__()
@@ -826,7 +970,129 @@ class LambdaCallback(Callback):
             self.on_dataloader_end = lambda logs: None
 
 
+class Summary(Callback):
+    '''调用torchinfo的summary
+    '''
+    def on_train_begin(self, logs=None):
+        from torchinfo import summary
+        print()
+        summary(self.model, input_data=next(iter(self.trainer.train_dataloader))[0])
+        print()
+
+
+def take_along_dim(input_tensor, indices, dim=None):
+    '''兼容部分低版本pytorch没有torch.take_along_dim
+    '''
+    if version.parse(torch.__version__) > version.parse('1.8.1'):
+        return torch.take_along_dim(input_tensor, indices, dim)
+    else:
+        # 该逻辑仅在少量数据上测试，如有bug，欢迎反馈
+        if dim is None:
+            res = input_tensor.flatten()[indices]
+        else:
+            res = np.take_along_axis(input_tensor.cpu().numpy(), indices.cpu().numpy(), axis=dim)
+            res = torch.from_numpy(res).to(input_tensor.device)
+        # assert res.equal(torch.take_along_dim(input_tensor, indices, dim))
+        return res
+
+
+def torch_div(input, other, rounding_mode=None):
+    ''' torch.div兼容老版本
+    '''
+    if version.parse(torch.__version__) < version.parse('1.7.2'):
+        indices = input // other  # 兼容老版本
+    else:
+        indices = torch.div(input, other, rounding_mode=rounding_mode)  # 行索引
+    return indices
+
+
+def softmax(x, axis=-1):
+    """numpy版softmax
+    """
+    x = x - x.max(axis=axis, keepdims=True)
+    x = np.exp(x)
+    return x / x.sum(axis=axis, keepdims=True)
+
+
+def search_layer(model, layer_name, retrun_first=True):
+    '''根据layer_name搜索并返回参数/参数list
+    '''
+    return_list = []
+    for name, param in model.named_parameters():
+        if param.requires_grad and layer_name in name:
+            return_list.append(param)
+    if len(return_list) == 0:
+        return None
+    if retrun_first:
+        return return_list[0]
+    else:
+        return return_list
+
+
+class ListDataset(Dataset):
+    '''数据是List格式Dataset，支持传入file_path或者外部已读入的data(List格式)
+
+    :param file_path: str, 待读取的文件的路径，若无可以为None
+    :param data: List[Any], list格式的数据，和file_path至少有一个不为None
+    '''
+    def __init__(self, file_path=None, data=None, **kwargs):
+        self.kwargs = kwargs
+        if isinstance(file_path, (str, tuple, list)):
+            self.data = self.load_data(file_path)
+        elif isinstance(data, list):
+            self.data = data
+        else:
+            raise ValueError('The input args shall be str format file_path / list format dataset')
+
+    def __len__(self):
+        return len(self.data)
+
+    def __getitem__(self, index):
+        return self.data[index]
+
+    @staticmethod
+    def load_data(file_path):
+        return file_path
+
+
+class IterDataset(IterableDataset):
+    '''流式读取文件，用于大数据量、多小文件使用时候需要注意steps_per_epoch != None
+
+    :param file_path: str, 待读取的文件的路径，若无可以为None
+    '''
+    def __init__(self, file_path=None, **kwargs):
+        self.kwargs = kwargs
+        if isinstance(file_path, (str, tuple, list)):
+            self.file_path = file_path
+        else:
+            raise ValueError('The input args shall be str format file_path / list format dataset')
+    
+    def __iter__(self):
+        return self.load_data(self.file_path)
+
+    @staticmethod
+    def load_data(file_path, verbose=0):
+        if isinstance(file_path, (tuple, list)):
+            for file in file_path:
+                if verbose != 0:
+                    print("Load data: ", file)
+                with open(file, 'r') as file_obj:
+                    for line in file_obj:
+                        yield line
+        elif isinstance(file_path, str):
+            with open(file_path, 'r') as file_obj:
+                for line in file_obj:
+                    yield line
+
+
 def metric_mapping(metric, func, y_pred, y_true):
+    '''metric的计算
+
+    :param metric: str, 自带metrics的名称
+    :param func: function, 透传的用户自定的计算指标的函数
+    :param y_pred: torch.Tensor, 样本的预测结果
+    :param y_true: torch.Tensor, 样本的真实结果
+    '''
     # 自定义metrics
     if inspect.isfunction(func):
         metric_res = func(y_pred, y_true)
@@ -878,6 +1144,8 @@ def metric_mapping(metric, func, y_pred, y_true):
 
 def seed_everything(seed=None):
     '''固定seed
+    
+    :param seed: int, 随机种子
     '''
     max_seed_value = np.iinfo(np.uint32).max
     min_seed_value = np.iinfo(np.uint32).min
@@ -891,4 +1159,6 @@ def seed_everything(seed=None):
     torch.manual_seed(seed)
     torch.cuda.manual_seed(seed)
     torch.cuda.manual_seed_all(seed)
+    torch.backends.cudnn.benchmark = False
+    torch.backends.cudnn.deterministic = True
     return seed
\ No newline at end of file
